# w.c.s. - web application for online forms
# Copyright (C) 2005-2023  Entr'ouvert
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import datetime
import uuid

from wcs import wf
from wcs.qommon import _
from wcs.qommon.form import SingleSelectWidget, StringWidget, WidgetList
from wcs.qommon.humantime import humanduration2seconds, seconds2humanduration, timewords
from wcs.qommon.xml_storage import XmlStorableObject
from wcs.testdef import TestError
from wcs.wf.backoffice_fields import SetBackofficeFieldRowWidget, SetBackofficeFieldsTableWidget
from wcs.wf.profile import FieldNode


class WorkflowTestError(TestError):
    pass


def get_test_action_options():
    return [(x.key, x.label, x.key) for x in WorkflowTestAction.__subclasses__()]


def get_test_action_class_by_type(action_type):
    for action_class in WorkflowTestAction.__subclasses__():
        if action_class.key == action_type:
            return action_class

    raise KeyError


class WorkflowTests(XmlStorableObject):
    _names = 'workflow_tests'
    xml_root_node = 'workflow_tests'
    testdef_id = None
    actions = None

    XML_NODES = [
        ('testdef_id', 'int'),
        ('actions', 'actions'),
    ]

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.actions = []

    def run(self, formdata, agent_user):
        # mock methods so nothing is stored
        formdata.record_workflow_event = lambda *args, **kwargs: None
        formdata.record_workflow_action = lambda *args, **kwargs: None
        formdata.store = lambda *args, **kwargs: None

        formdata.workflow_tests_data = {
            'now': formdata.receipt_time,
            'sent_emails': [],
        }

        status = None
        for action in self.actions:
            if status != formdata.get_status():
                formdata.perform_workflow(workflow_test=True)
                status = formdata.get_status()

            try:
                action.perform(formdata, agent_user)
            except WorkflowTestError as e:
                e.action_uuid = action.uuid
                e.details.append(_('Form status when error occured: %s') % status.name)
                raise e

    def get_new_action_id(self):
        if not self.actions:
            return '1'

        return str(int(max(x.id for x in self.actions)) + 1)

    def add_action(self, action_class):
        self.actions.append(action_class(id=self.get_new_action_id()))

    def export_actions_to_xml(self, element, attribute_name, **kwargs):
        for action in self.actions:
            element.append(action.export_to_xml())

    def import_actions_from_xml(self, element, **kwargs):
        actions = []
        for sub in element.findall('test-action'):
            key = sub.findtext('key')

            try:
                klass = get_test_action_class_by_type(key)
            except KeyError:
                continue

            actions.append(klass.import_from_xml_tree(sub))

        return actions


class WorkflowTestAction(XmlStorableObject):
    xml_root_node = 'test-action'
    _names = 'test-action'
    uuid = None

    optional_fields = []

    XML_NODES = [
        ('id', 'str'),
        ('uuid', 'str'),
        ('key', 'str'),
    ]

    def __init__(self, **kwargs):
        self.uuid = str(uuid.uuid4())

        allowed_key = {x[0] for x in self.XML_NODES}
        for k, v in kwargs.items():
            if k in allowed_key:
                setattr(self, k, v)

    def __str__(self):
        return str(self.label)

    def render_as_line(self):
        for field, dummy in self.XML_NODES:
            if field not in self.optional_fields and not getattr(self, field):
                return 'not configured'

        return self.details_label


class ButtonClick(WorkflowTestAction):
    label = _('Simulate click on action button')

    key = 'button-click'
    button_name = None

    XML_NODES = WorkflowTestAction.XML_NODES + [
        ('button_name', 'str'),
    ]

    @property
    def details_label(self):
        return 'Click on "%s"' % self.button_name

    def perform(self, formdata, user):
        status = formdata.get_status()
        form = status.get_action_form(formdata, user)
        if not form or not any(
            button_widget := x for x in form.submit_widgets if x.label == self.button_name
        ):
            raise WorkflowTestError(_('Button "%s" is not displayed.') % self.button_name)

        form.get_submit = lambda: button_widget.name
        status.handle_form(form, formdata, user, check_replay=False)

    def fill_admin_form(self, form, formdef):
        possible_button_names = set()
        for item in formdef.workflow.get_all_items():
            if isinstance(item, wf.choice.ChoiceWorkflowStatusItem) and item.status:
                possible_button_names.add(item.label)

        possible_button_names = sorted(possible_button_names)

        value = self.button_name
        if value and value not in possible_button_names:
            value = '%s (%s)' % (value, _('not available'))
            possible_button_names.append(value)

        form.add(
            SingleSelectWidget,
            'button_name',
            title=_('Button name'),
            options=possible_button_names,
            required=True,
            value=value,
        )


class AssertStatus(WorkflowTestAction):
    label = _('Assert form status')

    key = 'assert-status'
    status_name = None

    XML_NODES = WorkflowTestAction.XML_NODES + [
        ('status_name', 'str'),
    ]

    @property
    def details_label(self):
        return 'Status is "%s"' % self.status_name

    def perform(self, formdata, user):
        status = formdata.get_status()
        if status.name != self.status_name:
            raise WorkflowTestError(
                _('Form should be in status "%(expected_status)s" but is in status "%(status)s".')
                % {'expected_status': self.status_name, 'status': status.name}
            )

    def fill_admin_form(self, form, formdef):
        possible_statuses = [x.name for x in formdef.workflow.possible_status]

        value = self.status_name
        if value and value not in possible_statuses:
            value = '%s (%s)' % (value, _('not available'))
            possible_statuses.append(value)

        form.add(
            SingleSelectWidget,
            'status_name',
            title=_('Status name'),
            options=possible_statuses,
            required=True,
            value=self.status_name,
        )


class AssertEmail(WorkflowTestAction):
    label = _('Assert email is sent')

    key = 'assert-email'
    subject_strings = None
    body_strings = None

    optional_fields = ['subject_strings', 'body_strings']

    XML_NODES = WorkflowTestAction.XML_NODES + [
        ('subject_strings', 'str_list'),
        ('body_strings', 'str_list'),
    ]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.subject_strings = self.subject_strings or []
        self.body_strings = self.body_strings or []

    @property
    def details_label(self):
        return ''

    def perform(self, formdata, user):
        try:
            email = formdata.workflow_tests_data['sent_emails'].pop(0)
        except IndexError:
            raise WorkflowTestError(_('No email was sent.'))

        for subject in self.subject_strings:
            details = [_('Email subject: %s') % email.email_msg.subject]
            if subject not in email.email_msg.subject:
                raise WorkflowTestError(_('Email subject does not contain "%s".') % subject, details=details)

        for body in self.body_strings:
            details = [_('Email body: %s') % email.email_msg.body]
            if body not in email.email_msg.body:
                raise WorkflowTestError(_('Email body does not contain "%s".') % body, details=details)

    def fill_admin_form(self, form, formdef):
        form.add(
            WidgetList,
            'subject_strings',
            element_type=StringWidget,
            title=_('Subject must contain'),
            value=self.subject_strings,
            add_element_label=_('Add string'),
            element_kwargs={'render_br': False, 'size': 50},
        )
        form.add(
            WidgetList,
            'body_strings',
            element_type=StringWidget,
            title=_('Body must contain'),
            value=self.body_strings,
            add_element_label=_('Add string'),
            element_kwargs={'render_br': False, 'size': 50},
        )


class SkipTime(WorkflowTestAction):
    label = _('Move forward in time')

    key = 'skip-time'
    seconds = None

    XML_NODES = WorkflowTestAction.XML_NODES + [
        ('seconds', 'int'),
    ]

    @property
    def details_label(self):
        return seconds2humanduration(self.seconds)

    def rewind(self, formdata):
        def rewind_time(timestamp):
            return timestamp - datetime.timedelta(seconds=self.seconds)

        formdata.receipt_time = rewind_time(formdata.receipt_time)
        formdata.evolution[-1].time = rewind_time(formdata.evolution[-1].time)

    def perform(self, formdata, user):
        self.rewind(formdata)

        jump_actions = []
        status = formdata.get_status()
        for item in status.items:
            if hasattr(item, 'has_valid_timeout') and item.has_valid_timeout():
                jump_actions.append(item)

        delay = wf.jump.get_min_jumps_delay(status.items)

        if formdata.last_update_time > formdata.workflow_tests_data['now'] - datetime.timedelta(
            seconds=delay
        ):
            return

        for jump_action in jump_actions:
            if jump_action.check_condition(formdata):
                wf.jump.jump_and_perform(formdata, jump_action)
                break

    def fill_admin_form(self, form, formdef):
        form.add(
            StringWidget,
            'seconds',
            title=_('Value'),
            value=seconds2humanduration(self.seconds),
            hint=_('ex.: 1 day 12 hours. Usable units of time: %(variables)s.')
            % {'variables': ','.join(timewords())},
        )

    def seconds_parse(self, value):
        if not value:
            return value
        try:
            return humanduration2seconds(value)
        except ValueError:
            return None


class AssertBackofficeFieldRowWidget(SetBackofficeFieldRowWidget):
    value_widget = StringWidget
    value_placeholder = None


class AssertBackofficeFieldsTableWidget(SetBackofficeFieldsTableWidget):
    element_type = AssertBackofficeFieldRowWidget


class AssertBackofficeFieldValues(WorkflowTestAction):
    label = _('Assert backoffice field values')

    key = 'assert-backoffice-field'
    fields = []

    XML_NODES = WorkflowTestAction.XML_NODES + [
        ('fields', 'fields'),
    ]

    @property
    def details_label(self):
        return ''

    def perform(self, formdata, user):
        for field_dict in self.fields:
            field_id = field_dict['field_id']
            expected_value = field_dict['value']
            formdata_value = formdata.data.get(field_id)

            if formdata_value != expected_value:
                fields = [x for x in formdata.formdef.workflow.get_backoffice_fields() if x.id == field_id]
                if not fields:
                    raise WorkflowTestError(
                        _('Field %(field_id)s not found (expected value "%(value)s").')
                        % {
                            'field_id': field_id,
                            'value': expected_value,
                        }
                    )

                field = fields[0]
                raise WorkflowTestError(
                    _(
                        'Wrong value for backoffice field "%(field)s" (expected "%(expected_value)s", got "%(value)s").'
                    )
                    % {
                        'field': field.label,
                        'value': formdata_value,
                        'expected_value': expected_value,
                    }
                )

    def fill_admin_form(self, form, formdef):
        form.add(
            AssertBackofficeFieldsTableWidget,
            'fields',
            value_widget_class=StringWidget,
            value=self.fields,
            workflow=formdef.workflow,
        )

    def export_fields_to_xml(self, element, attribute_name, **kwargs):
        for field in self.fields:
            element.append(FieldNode(field).export_to_xml(include_id=True))

    def import_fields_from_xml(self, element, **kwargs):
        fields = []
        for field_xml_node in element.findall('field'):
            field_node = FieldNode()
            field_node.init_with_xml(field_xml_node, include_id=True, snapshot=None)
            fields.append(field_node.as_dict())

        return fields
